"""Generic FIX Order single module."""
import re
from datetime import datetime
from math import isfinite, nan

from asyncfix import FIXMessage, FMsg, FTag
from asyncfix.errors import FIXError

from .common import FExecType, FOrdSide, FOrdStatus, FOrdType

RE_CLORD_ROOT = re.compile(r"^(.+)--(\d+)$", re.MULTILINE)


class FIXNewOrderSingle:
    """Generic FIXNewOrderSingle wrapper.

    Attributes:
        clord_id: current order ClOrdID
        orig_clord_id: order OrigClOrdID (when canceling / replacing)
        order_id: executed order id
        ticker: user order ticker
        side: order side
        price: order price
        qty: order quantity
        leaves_qty: order remaining qty
        cum_qty: order filled qty
        avg_px: average fill price
        ord_type: order type
        account: order account
        status: current order status
        target_price: order target execution price
    """

    def __init__(
        self,
        clord_id: str,
        cl_ticker: str,
        side: FOrdSide | str,
        price: float,
        qty: float,
        ord_type: FOrdType | str = FOrdType.LIMIT,
        account: str | dict = "000000",
        target_price: float | None = None,
    ):
        """Initialize order.

        Args:
            clord_id: root ClOrdID
            cl_ticker: client ticker (can by any accepted by user's OMS)
            side: order side
            price: order initial price (current price, changes if replaced)
            qty: order quantity
            ord_type: order type
            account: order account (optional)
            target_price: order initial target execution price (useful for slippage)
        """
        assert clord_id, "empty clord_id"

        self.clord_id = clord_id
        self.orig_clord_id = None
        self.order_id = None
        self.ticker = cl_ticker
        self.side = side
        self.price = price
        self.qty = qty
        self.leaves_qty = 0.0
        self.cum_qty = 0.0
        self.avg_px = nan
        self.ord_type = ord_type
        self.account = account
        self._clord_id_cnt = 0
        self.status: FOrdStatus = FOrdStatus.CREATED
        self.target_price = target_price if target_price is not None else price

    def __repr__(self):
        """Repr."""
        return (
            f"FIXNewOrderSingle({self.status.name}, clord={self.clord_id},"
            f" ticker={self.ticker}, px={self.price}, qty={self.qty},"
            f" leavesqty={self.leaves_qty}, cumqty={self.cum_qty})"
        )

    def clord_next(self) -> str:
        """New ClOrdID for current order management."""
        self._clord_id_cnt += 1
        return f"{self.clord_id_root}--{self._clord_id_cnt}"

    @staticmethod
    def clord_root(clord_id: str) -> str:
        """Order ClOrdID root, as given at initialization.

        Args:
            clord_id: current order one of the clord_next()

        Returns:
            string
        """
        match = RE_CLORD_ROOT.match(clord_id)
        if match:
            return match[1]
        else:
            return clord_id

    @property
    def clord_id_root(self) -> str:
        """Current order ClOrdID root."""
        return self.clord_root(self.clord_id)

    @staticmethod
    def current_datetime():
        """Date for TransactTime field."""
        return datetime.utcnow().strftime("%Y%m%d-%H:%M:%S.%f")[:-3]

    def new_req(self) -> FIXMessage:
        """Creates NewOrderSingle message."""
        assert (
            self.status == FOrdStatus.CREATED
        ), "new_req() must send only just created orders"

        o = FIXMessage(FMsg.NEWORDERSINGLE)
        self.clord_id = self.clord_next()
        o[FTag.ClOrdID] = self.clord_id

        # setting instrument identification fields (may vary for different FIX brokers)
        self.set_instrument(o)

        # setting account field (may vary for different FIX brokers)
        self.set_account(o)
        o[FTag.OrdType] = self.ord_type
        o[FTag.Side] = self.side
        o[FTag.TransactTime] = self.current_datetime()

        # setting price and qty (typically tick size rounding, etc)
        self.set_price_qty(o, self.price, self.qty)

        self.status = FOrdStatus.PENDING_NEW

        return o

    def cancel_req(self) -> FIXMessage:
        """Creates order cancel request.

        Raises:
            FIXError: if order can't be canceled
        """
        if not self.can_cancel():
            raise FIXError(f"{self} Fix order is not allowed for cancel")

        assert not self.orig_clord_id
        self.orig_clord_id = self.clord_id
        self.clord_id = self.clord_next()

        cxl_req_msg = FIXMessage(FMsg.ORDERCANCELREQUEST)
        cxl_req_msg[11] = self.clord_id
        cxl_req_msg[38] = self.qty
        cxl_req_msg[41] = self.orig_clord_id
        self.set_instrument(cxl_req_msg)
        cxl_req_msg[FTag.Side] = self.side
        cxl_req_msg[FTag.TransactTime] = self.current_datetime()

        self.status = FOrdStatus.PENDING_CANCEL

        return cxl_req_msg

    def replace_req(self, price: float = nan, qty: float = nan) -> FIXMessage:
        """Creates order replace request.

        Args:
            price: alternative price
            qty: alternative qty

        Returns:
            message

        Raises:
            FIXError: if order can't be replaced or price/qty unchanged
        """
        if not self.can_replace():
            raise FIXError("Order cannot be replaced")

        if price is None or not isfinite(price) or price == self.price:
            price = self.price
        if qty is None or not isfinite(qty) or qty == self.qty or qty == 0:
            qty = self.qty

        if price == self.price and qty == self.qty:
            raise FIXError("no price / qty change in replace_req")

        assert not self.orig_clord_id
        self.orig_clord_id = self.clord_id
        self.clord_id = self.clord_next()

        m = FIXMessage(FMsg.ORDERCANCELREPLACEREQUEST)
        m[FTag.ClOrdID] = self.clord_id
        m[FTag.OrigClOrdID] = self.orig_clord_id
        m[FTag.OrdType] = self.ord_type
        self.set_instrument(m)
        self.set_price_qty(m, price, qty)
        m[FTag.Side] = self.side
        m[FTag.TransactTime] = self.current_datetime()

        self.status = FOrdStatus.PENDING_REPLACE

        return m

    def set_instrument(self, ord_msg: FIXMessage):
        """Set order instrument definition (override this in child).

        Args:
            ord_msg: new or replaced order
        """
        # Simply populate symbol, in read life FIX counterparty may require populate
        #  more fields (override this method in child class)
        ord_msg[FTag.Symbol] = self.ticker

    def set_account(self, ord_msg: FIXMessage):
        """Set account definition (override this in child).

        Args:
            ord_msg: new or replaced order
        """
        # Simplistic account setting (you must override this method in child class)
        assert isinstance(self.account, str), "Account expected a string"
        ord_msg[FTag.Account] = self.account

    def set_price_qty(self, ord_msg: FIXMessage, price: float, qty: float):
        """Set order price and qty definition (override this in child).

        This method handles custom price/qty rounding/decimal formatting, or maybe
        conditional presence of two fields based on order type

        Args:
            ord_msg: new or replaced order
            price: new order price (unformatted / unrounded)
            qty: new order qty (unformatted / unrounded)
        """
        ord_msg[FTag.Price] = price
        ord_msg[FTag.OrderQty] = qty

    @staticmethod
    def change_status(
        status: FOrdStatus,
        fix_msg_type: FMsg,
        msg_exec_type: FExecType,
        msg_status: FOrdStatus,
        raise_on_err: bool = True,
    ) -> FOrdStatus | None:
        """FIX Order State transition algo.

        :param status: current order status
        :param fix_msg_type: incoming/or requesting order type,  these are supported:
                '8' - execution report,
                '9' - Order Cancel reject,
                'F' - Order cancel request (if possible to cancel current order)
                'G' -  Order replace request (if possible to replace current order)
        :param msg_exec_type: (only for execution report), for other should be 0
        :param msg_status: new fix msg order status, or required status
        :return: FOrdStatus if state transition is possible,
                 None - if transition is valid, but need to wait for a good state
                 raises FIXError - when transition is invalid
        """
        status_transitions = {}

        if fix_msg_type == FMsg.EXECUTIONREPORT:
            status_transitions = {
                None: {None: FIXError},
                # key {initial status}: {
                #    msg_status: <transition>,
                #       <transition>: None - ignore, True - transit, FIXError - raise
                #
                #      REJECTED: True,     #  this is allowed status transition
                #      PENDING_NEW: None,  #  transition allowed but no status change
                #      CREATED: FIXError,  #  error transition
                #      None:  [None, True, FIXError] # default transition
                #  }
                FOrdStatus.CREATED: {
                    FOrdStatus.PENDING_NEW: True,
                    FOrdStatus.REJECTED: True,
                    None: FIXError,
                },
                FOrdStatus.PENDING_NEW: {
                    FOrdStatus.REJECTED: True,
                    FOrdStatus.NEW: True,
                    FOrdStatus.FILLED: True,
                    FOrdStatus.PARTIALLY_FILLED: True,
                    FOrdStatus.CANCELED: True,
                    FOrdStatus.SUSPENDED: True,
                    None: FIXError,
                },
                FOrdStatus.NEW: {
                    FOrdStatus.NEW: None,
                    FOrdStatus.PENDING_NEW: FIXError,
                    FOrdStatus.CREATED: FIXError,
                    FOrdStatus.ACCEPTED_FOR_BIDDING: FIXError,
                    None: True,
                },
                FOrdStatus.FILLED: {
                    None: None,
                },
                FOrdStatus.CANCELED: {
                    None: None,
                },
                FOrdStatus.REJECTED: {
                    None: None,
                },
                FOrdStatus.EXPIRED: {
                    None: None,
                },
                FOrdStatus.SUSPENDED: {
                    FOrdStatus.NEW: True,
                    FOrdStatus.PARTIALLY_FILLED: True,
                    FOrdStatus.CANCELED: True,
                    FOrdStatus.SUSPENDED: None,
                    None: FIXError,
                },
                FOrdStatus.PARTIALLY_FILLED: {
                    FOrdStatus.FILLED: True,
                    FOrdStatus.PARTIALLY_FILLED: True,
                    FOrdStatus.PENDING_REPLACE: True,
                    FOrdStatus.PENDING_CANCEL: True,
                    FOrdStatus.CANCELED: True,
                    FOrdStatus.EXPIRED: True,
                    FOrdStatus.SUSPENDED: True,
                    FOrdStatus.STOPPED: True,
                    None: FIXError,
                },
                FOrdStatus.PENDING_CANCEL: {
                    FOrdStatus.CANCELED: True,
                    FOrdStatus.CREATED: FIXError,
                    None: None,
                },
                FOrdStatus.PENDING_REPLACE: {
                    "exec_type": {
                        FExecType.REPLACED: {
                            FOrdStatus.NEW: True,
                            FOrdStatus.PARTIALLY_FILLED: True,
                            FOrdStatus.FILLED: True,
                            FOrdStatus.CANCELED: True,
                            None: FIXError,
                        },
                        None: {
                            FOrdStatus.CREATED: FIXError,
                            FOrdStatus.ACCEPTED_FOR_BIDDING: FIXError,
                            None: None,
                        },
                    },
                },
            }

        elif fix_msg_type == FMsg.ORDERCANCELREJECT:  # '9'
            status_transitions = {
                None: {
                    FOrdStatus.CREATED: FIXError,
                    FOrdStatus.ACCEPTED_FOR_BIDDING: FIXError,
                    None: True,
                }
            }
        elif (
            fix_msg_type == FMsg.ORDERCANCELREQUEST
            or fix_msg_type == FMsg.ORDERCANCELREPLACEREQUEST
        ):
            status_transitions = {
                FOrdStatus.PENDING_CANCEL: {None: None},
                FOrdStatus.PENDING_REPLACE: {None: None},
                FOrdStatus.NEW: {None: True},
                FOrdStatus.SUSPENDED: {None: True},
                FOrdStatus.PARTIALLY_FILLED: {None: True},
                None: {None: FIXError},
            }

        if not status_transitions:
            raise FIXError(f"No status transition table for {fix_msg_type=}")

        s = status_transitions.get(status, status_transitions[None])
        if isinstance(s, dict) and "exec_type" in s:
            s = s["exec_type"].get(msg_exec_type, s["exec_type"][None])
        default = s[None]
        result = s.get(msg_status, default)

        if result is FIXError:
            # Invalid transition
            if raise_on_err:
                raise FIXError("FIX Order state transition error")
            else:
                return None
        else:
            if result is None:
                # Valid transition no status change
                return None
            else:
                # Do status change
                return msg_status

    def process_cancel_rej_report(self, m: FIXMessage) -> bool:
        """Processes incoming cancel reject report message."""
        if m.msg_type != FMsg.ORDERCANCELREJECT:
            raise FIXError("incorrect message type")

        order_status = m[FTag.OrdStatus]

        new_status = self.change_status(
            self.status, m.msg_type, 0, order_status, raise_on_err=False
        )
        #
        if order_status == FOrdStatus.REJECTED:
            # Very weird (emergency) case, because the ClOrdId does not exist
            #   Let's set order inactive
            self.leaves_qty = 0

        if new_status is not None:
            self.status = new_status
            return True
        else:
            return False

    def process_execution_report(self, m: FIXMessage) -> bool:
        """Processes incoming execution report for an order.

        Raises:
            FIXError: if ClOrdID mismatch
        """
        if m.msg_type != FMsg.EXECUTIONREPORT:
            raise FIXError("incorrect message type")

        clord_id = m[FTag.ClOrdID]
        cum_qty = float(m[FTag.CumQty])
        order_status = m[FTag.OrdStatus]

        if clord_id != self.clord_id and clord_id != self.orig_clord_id:
            raise FIXError("orig_clord_id mismatch")

        exec_type = m[FTag.ExecType]
        leaves_qty = float(m[FTag.LeavesQty])

        new_status = FIXNewOrderSingle.change_status(
            self.status, m.msg_type, exec_type, order_status, raise_on_err=False
        )

        self.order_id = m[FTag.OrderID]
        self.leaves_qty = leaves_qty
        self.cum_qty = cum_qty
        self.avg_px = float(m[FTag.AvgPx])

        if exec_type == FExecType.REPLACED:
            # Order has been successfully replaced
            new_price = m.get(FTag.Price, None)
            if new_price is not None:
                # Price may not be in execution report, it's not an error
                self.price = float(new_price)

            order_qty = m.get(FTag.OrderQty, None)
            if order_qty is not None:
                # Qty may not be in execution report, it's not an error
                self.qty = float(order_qty)

            # Clearing OrigOrdId to allow subsequent order changes
            self.orig_clord_id = None

        if new_status:
            self.status = FOrdStatus(new_status)
            return True
        else:
            return False

    def is_finished(self) -> bool:
        """Check if order is in terminal state (no subsequent changes expected)."""
        return (
            self.status == FOrdStatus.FILLED
            or self.status == FOrdStatus.CANCELED
            or self.status == FOrdStatus.REJECTED
            or self.status == FOrdStatus.EXPIRED
        )

    def can_cancel(self) -> bool:
        """Check if order can be canceled from its current state."""
        return (
            FIXNewOrderSingle.change_status(
                self.status,
                FMsg.ORDERCANCELREQUEST,
                0,  # Exec Type - omitted!
                FOrdStatus.PENDING_CANCEL,
                raise_on_err=False,
            )
            is not None
        )

    def can_replace(self) -> bool:
        """Check if order can be replaced from its current state."""
        return (
            FIXNewOrderSingle.change_status(
                self.status,
                FMsg.ORDERCANCELREPLACEREQUEST,
                0,  # Exec Type - omitted!
                FOrdStatus.PENDING_REPLACE,
                raise_on_err=False,
            )
            is not None
        )



"""FIX Schema validation module."""
from __future__ import annotations

import dataclasses
import datetime as dtm
import re
import warnings
import xml.etree.ElementTree as ET
from math import isfinite

from asyncfix import FIXMessage, FTag
from asyncfix.errors import FIXMessageError
from asyncfix.message import FIXContainer


@dataclasses.dataclass
class SchemaField:
    """FIX Field schema."""

    tag: str
    name: str
    ftype: str
    values: dict[str, str] = dataclasses.field(default_factory=dict)

    def __hash__(self):
        """Hash by Field.name."""
        return hash(self.name)

    def __eq__(self, o):
        """Equality by SchemaField.name or tag."""
        if isinstance(o, SchemaField):
            return self.name == o.name
        elif isinstance(o, (str, int)):
            try:
                return self.tag == str(int(o))
            except Exception:
                return self.name == o
        else:
            return False

    def validate_value(self, value: str) -> bool:
        """Validate tag value based on schema settings.

        Args:
            value: tag value

        Returns: True - if passed

        Raises:
            FIXMessageError: raised if validation failed
        """
        assert isinstance(value, str), "value must be a string"
        assert value, "empty value"

        if self.values:
            if value not in self.values:
                raise FIXMessageError(
                    f"Field={self.name} value expected to be one of the"
                    f" {list(self.values.keys())}, got ({value=})"
                )
            return True
        else:
            t = self.ftype.upper()
            err = None
            if t == "INT":
                err = SchemaField._validate_value_number(value, int)
            elif t in ["SEQNUM", "NUMINGROUP"]:
                err = SchemaField._validate_value_number(
                    value,
                    int,
                    no_zero=True,
                    no_negative=True,
                )
            elif t == "DAYOFMONTH":
                err = SchemaField._validate_value_number(value, int, num_range=(1, 31))
            elif t in {"FLOAT", "QTY", "PRICE", "PRICEOFFSET", "AMT", "PERCENTAGE"}:
                err = SchemaField._validate_value_number(
                    value, float, no_nonfinite=True
                )
            elif t in {"STRING", "MULTIPLESTRINGVALUE"}:
                err = SchemaField._validate_value_str(value)
            elif t in {"CHAR"}:
                err = SchemaField._validate_value_str(value, max_len=1)
            elif t in {"BOOLEAN"}:
                err = SchemaField._validate_value_str(
                    value, max_len=1, subset={"Y", "N"}
                )
            elif t in {"COUNTRY"}:
                err = SchemaField._validate_value_str(value, max_len=2, alpha_num=True)
            elif t in {"CURRENCY"}:
                err = SchemaField._validate_value_str(value, max_len=3, alpha_num=True)
            elif t in {"EXCHANGE"}:
                err = SchemaField._validate_value_str(value, max_len=4, alpha_num=True)
            elif t in {"LOCALMKTDATE", "UTCDATEONLY"}:
                err = SchemaField._validate_value_datetime(value, "%Y%m%d")
            elif t in {"UTCTIMESTAMP"}:
                err = SchemaField._validate_value_datetime(value, "%Y%m%d-%H:%M:%S")
            elif t in {"UTCTIMEONLY"}:
                err = SchemaField._validate_value_datetime(value, "%H:%M:%S")
            elif t == "MONTHYEAR":
                err = SchemaField._validate_value_monthyear(value)
            elif t in {"DATA", "LENGTH"}:
                # just hoping the data is ok
                err = None
            else:
                warnings.warn(f"Unsupported datatype: {t} for field={self}")

            err = self._validate_special_cases(value, err)

            if err:
                raise FIXMessageError(f"{self} validation error (value={value}): {err}")

            return True

    def _validate_special_cases(self, value, prev_err):
        if self.tag == "16":
            # EndSeqNo
            if value == "0":
                prev_err = None
        return prev_err

    @staticmethod
    def _validate_value_datetime(value, format):
        if "." in value and "%S" in format and "%f" not in format:
            format = f"{format}.%f"

        try:
            dtm.datetime.strptime(value, format)
            return None  # all good
        except Exception as exc:
            return str(exc)

    @staticmethod
    def _validate_value_monthyear(value):
        """Special case for MonthYear type.

        String representing month of a year. An optional day of the month can be
          appended or an optional week code.
        Valid formats:
        YYYYMM
        YYYYMMDD
        YYYYMMWW.

        Valid values:
        YYYY = 0000-9999; MM = 01-12; DD = 01-31;
        WW = w1, w2, w3, w4, w5.
        """
        if "w" in value:
            week = value[-2:]
            if week not in {"w1", "w2", "w3", "w4", "w5"}:
                return f"YYYYMMWW [ww must be in w1,w2,w3,w4,w5], got {week}"
            format = "%Y%m"
            value = value[:-2]
            if len(value) != 6:
                return "remainder YYYYMM invalid"
        else:
            if len(value) == 6:
                format = "%Y%m"
            else:
                format = "%Y%m%d"

        return SchemaField._validate_value_datetime(value, format)

    @staticmethod
    def _validate_value_str(value, max_len=None, subset=None, alpha_num=False):
        assert value
        assert type(value) is str
        if "\x01" in value:
            return "Value contains SOH char"
        if "=" in value:
            return "Value contains `=` char"

        if max_len and len(value) > max_len:
            return "max legth exceeded"
        if subset and value not in subset:
            return f"out of subset: {subset}"
        if alpha_num and re.search(r"\W+", value):
            return "value contains non alphanumeric letters"

    @staticmethod
    def _validate_value_number(
        value: str,
        num_type: type,
        no_zero=False,
        no_negative=False,
        no_nonfinite=False,
        num_range=None,
    ) -> str | None:
        assert value
        assert num_type in (int, float)

        try:
            v = num_type(value)
            if no_zero and v == 0:
                raise ValueError("zero value")
            if no_negative and v < 0:
                raise ValueError("negative value")
            if no_nonfinite and not isfinite(float(v)):
                raise ValueError("not isfinite number")
            if num_range and not (v >= num_range[0] and v <= num_range[1]):
                raise ValueError(f"out of range {num_range}")
            # all good
            return None
        except ValueError as exc:
            return str(exc)

    def __str__(self):
        """As string: name|tag."""
        return f"{self.name}|{self.tag}"

    def __repr__(self):
        """Repr."""
        return f"SchemaField({str(self)}, type={self.ftype})"


class SchemaSet:
    """Generic schema set (base for component/group).

    Attributes:
        name: schema name
        field: schema field
        members: members
        required: required flag
    """

    def __init__(self, name: str, field: SchemaField | None = None):
        """Initialize.

        Args:
            name: name of abstract set
            field: field of abstract set

        Raises:
            ValueError: if not NUMINGROUP type or similar tag name
        """
        self.name: str = name
        self.field: SchemaField = field  # this is set for groups, with attached field
        if field:
            if not (
                ("No" in field.name or "Num" in field.name)
                and field.ftype in ["NUMINGROUP", "INT"]
            ):
                # warnings.warn('')
                raise ValueError(
                    "SchemaSet expected to have group field with NUMINGROUP|INT type,"
                    f" name contais No/Num, got {field}"
                )
        self.members: dict[SchemaField | SchemaSet, SchemaField | SchemaSet] = {}
        self.required: dict[SchemaField | SchemaSet, bool] = {}

    @property
    def tag(self) -> str:
        """Tag number of SchemaField.

        Returns:
            tag

        Raises:
            ValueError: raised then tag is not single field
        """
        if self.field:
            return self.field.tag
        else:
            raise ValueError(f"tag property is not supported for {self}")

    def keys(self) -> list[str]:
        """List of field names."""
        return [m.name for m in self.members]

    def add(self, field_or_set: SchemaField | SchemaSet, required: bool):
        """Add SchemaSet member.

        Args:
            field_or_set: field or SchemaSet
            required: required tag flag

        Raises:
            ValueError: unsupported field_or_set value
        """
        if isinstance(field_or_set, SchemaField):
            assert field_or_set not in self.members
        elif isinstance(field_or_set, SchemaSet):
            assert field_or_set.field, "field_or_set.field is empty, try to merge"
        else:
            raise ValueError(f"Unsupported field_or_set type, got {type(field_or_set)}")

        self.members[field_or_set] = field_or_set
        self.required[field_or_set] = required

    def merge(self, comp: SchemaSet):
        """Merge SchemaSet with another.

        Args:
            comp: SchemaSet
        """
        assert isinstance(comp, SchemaSet)

        for field, required in comp.members.items():
            self.add(field, comp.required[field])

    def __hash__(self):
        """Hash by field.name."""
        if self.field:
            return hash(self.field)
        else:
            return hash(self.name)

    def __eq__(self, o):
        """Equality."""
        if self.field:
            return self.field == o
        else:
            return self.name == o

    def __contains__(self, item: str | SchemaField) -> bool:
        """Check if SchemaSet contains field or name.

        Args:
            item: SchemaField or name (str)

        Returns:
            boolean

        Raises:
            FIXMessageError: incorrect item value
        """
        try:
            int(str(item))
            raise FIXMessageError("item looks like tag, use name or SchemaField")
        except ValueError:
            return item in self.members

    def __getitem__(self, item: str | SchemaField) -> SchemaField | SchemaSet:
        """Get field item by SchemaField or name.

        Args:
            item: SchemaField or name (str)

        Returns:
            SchemaField | SchemaSet

        Raises:
            FIXMessageError: incorrect item value
        """
        try:
            int(str(item))
            raise FIXMessageError("item looks like tag, use name or SchemaField")
        except ValueError:
            return self.members[item]


class SchemaGroup(SchemaSet):
    """SchemaGroup container.

    Attributes:
        field_required:
    """

    def __init__(self, field: SchemaField, required: bool):
        """Initialize.

        Args:
            field: SchemaField of group
            required: required flag
        """
        super().__init__(field.name, field)
        self.field_required = required

    def validate_group(self, groups: list[FIXContainer]):
        """Validate values of all tags in group.

        Args:
            groups: list of repeating group items

        Raises:
            FIXMessageError: validation failed
        """
        tag_order = {f.tag: i for i, f in enumerate(self.members.values())}
        tag_fields = {f.tag: f for f in self.members.values()}

        for fmsg in groups:
            has_first_tag = False
            prev_tag = -1

            for t, v in fmsg.items():
                if t not in tag_order:
                    raise FIXMessageError(
                        f"fixmessage={groups} contains unsupported tag for {self}"
                    )
                ord_idx = tag_order[t]

                if ord_idx == 0:
                    has_first_tag = True
                if prev_tag > ord_idx:
                    raise FIXMessageError(
                        f"fixmessage={groups} incorrect tag order {self}"
                    )

                field = tag_fields[t]

                if isinstance(field, SchemaField):
                    field.validate_value(v)
                else:
                    # Nested group!?
                    assert isinstance(field, SchemaGroup)
                    if not fmsg.is_group(t):
                        raise FIXMessageError(
                            f"fixmessage={groups}, tag={t} must be a group {self}"
                        )

                    # validate nested group
                    field.validate_group(fmsg.get_group_list(t))

                prev_tag = ord_idx

            if not has_first_tag:
                raise FIXMessageError(
                    f"fixmessage={groups} does not contain mandatory first tag {self}"
                )

            for st, sv in tag_fields.items():
                if isinstance(sv, SchemaField):
                    if sv.tag not in fmsg and self.required[sv]:
                        raise FIXMessageError(
                            f"fixmessage={groups} missing required field {repr(sv)}"
                        )

    def __repr__(self):
        """Repr."""
        members = [str(m) for m in self.members.keys()]
        return f"SchemaGroup({self.field.name}, {members})"


class SchemaComponent(SchemaSet):
    """SchemaComponent container."""

    def __init__(self, name: str):
        """Initialize.

        Args:
            name: component name
        """
        super().__init__(name)


class SchemaHeader(SchemaSet):
    """SchemaHeader container."""

    def __init__(self):
        """Initialize header."""
        super().__init__(name="Header")


class SchemaMessage(SchemaSet):
    """SchemaMessage container.

    Attributes:
        msg_type: msg_type value
        msg_cat: message category
    """

    def __init__(self, name: str, msg_type: str, msg_cat: str):
        """Initialize.

        Args:
            name: message name
            msg_type: message type
            msg_cat: message category
        """
        super().__init__(name)
        self.msg_type = msg_type
        self.msg_cat = msg_cat

    def __repr__(self):
        """Repr."""
        return (
            f"SchemaMessage(name={self.name}, type={self.msg_type}, cat={self.msg_cat})"
        )


class FIXSchema:
    """FIX schema validator."""

    def __init__(self, xml_or_path: ET.ElementTree | str):
        """Initialize.

        Args:
            xml_or_path: path to xml or xml.etree.ElementTree

        Raises:
            ValueError:
        """
        if isinstance(xml_or_path, str):
            xml_or_path = ET.parse(xml_or_path)

        assert isinstance(xml_or_path, ET.ElementTree)

        self._tag2field: dict[str, SchemaField] = {}
        self._field2tag: dict[str, SchemaField] = {}
        self._header: SchemaHeader = None
        self._components: dict[str, SchemaComponent] = {}
        self._messages: dict[str, SchemaMessage] = {}
        self._messages_types: dict[str, SchemaMessage] = {}
        self._header = {}
        self._types = set()

        self._parse(xml_or_path.getroot())

    def _parse_msg_set(self, component, element):
        has_circular_refs = False
        for val in element:
            assert val.tag in ["field", "group", "component"]

            if val.tag == "field":
                field = self._field2tag[val.attrib["name"]]
                component.add(field, val.attrib["required"].upper() == "Y")
            elif val.tag == "component":
                if val.attrib["name"] not in self._components:
                    # We have circular reference, component was referenced before
                    #    described in XML data, just skip and rerun later
                    has_circular_refs = True
                    continue
                ref_component = self._components[val.attrib["name"]]
                component.merge(ref_component)
            elif val.tag == "group":
                g = self._parse_group(val)
                if g is None:
                    # Group also refers to other component, postpone it
                    has_circular_refs = True
                    continue
                component.add(g, g.required)

        if has_circular_refs:
            return None
        else:
            return component

    def _parse_group(self, element: ET.Element) -> SchemaGroup | None:
        assert self._field2tag, "parse fields first!"
        assert element.tag == "group"

        el_name = element.attrib["name"]

        gfield = self._field2tag[el_name]
        group = self._parse_msg_set(
            SchemaGroup(gfield, element.attrib["required"].upper() == "Y"),
            element,
        )
        # assert group, f'group parse erorr, {element.attrib}'
        return group

    def _parse_component(self, element: ET.Element) -> SchemaComponent | None:
        assert self._field2tag, "parse fields first!"
        assert element.tag == "component"

        el_name = element.attrib["name"]

        assert el_name not in self._components, "Duplicate component name or double run"
        component = self._parse_msg_set(SchemaComponent(el_name), element)

        if component:
            self._components[el_name] = component
            return component
        else:
            return None

    def _parse_message(self, element: ET.Element) -> SchemaMessage | None:
        assert self._field2tag, "parse fields first!"
        assert element.tag == "message"

        el_name = element.attrib["name"]

        assert el_name not in self._messages, "Duplicate message?"

        message = self._parse_msg_set(
            SchemaMessage(
                el_name,
                msg_type=element.attrib["msgtype"],
                msg_cat=element.attrib["msgcat"],
            ),
            element,
        )
        assert message, "Message probably refers to circular refs in comp or groups"

        self._messages[el_name] = message
        self._messages_types[message.msg_type] = message
        return message

    def _parse_header(self, element: ET.Element):
        assert self._field2tag, "parse fields first!"
        assert element.tag == "header"

        self._header = self._parse_msg_set(SchemaHeader(), element)

    def _parse_field(self, element: ET.Element):
        assert element.tag == "field"

        f = SchemaField(
            tag=element.attrib["number"],
            name=element.attrib["name"],
            ftype=element.attrib["type"],
        )
        if len(element) > 0:
            for val in element:
                assert val.tag == "value"
                f.values[val.attrib["enum"]] = val.attrib["description"]

        self._types.add(f.ftype)
        self._tag2field[f.tag] = f
        self._field2tag[f.name] = f

    def _parse(self, root: ET.Element):
        assert not self._field2tag, "already parsed"

        for element in root.find("fields"):
            self._parse_field(element)

        self._parse_header(root.find("header"))

        all_components = [e for e in root.find("components")]
        full_count = len(all_components)
        prev_cnt = len(all_components)
        while all_components:
            i = 0
            while i < len(all_components):
                if self._parse_component(all_components[i]):
                    # component was parsed
                    del all_components[i]
                else:
                    i += 1

            if all_components:
                if len(all_components) == prev_cnt:
                    failed_comp = [c.attrib["name"] for c in all_components]
                    raise RuntimeError(
                        "Failed to resolve component circular reference."
                        f"Failed components: {failed_comp}"
                    )
                else:
                    prev_cnt = len(all_components)
        assert full_count == len(self._components), "Component count mismatch"

        n_msg = 0
        for element in root.find("messages"):
            self._parse_message(element)
            n_msg += 1

        assert n_msg == len(self._messages), "Message count mismatch"

    def _validate_header(self, msg: FIXMessage):
        schema_fields = set()
        schema_msg = self._header
        for fname, req in schema_msg.required.items():
            f = schema_msg[fname]
            schema_fields.add(fname)

            if req:
                if isinstance(f, SchemaField):
                    if f.tag not in msg:
                        raise FIXMessageError(f"Missing required field={repr(f)}")
                    f_val = msg[f.tag]
                    f.validate_value(f_val)

    def validate(self, msg: FIXMessage) -> bool:
        """Validates generic FIXMessage based on schema.

        Args:
            msg: generic FIXMessage

        Returns:
            True - if ok

        Raises:
            FIXMessageError: raises on invalid message
        """
        if msg.msg_type not in self._messages_types:
            raise FIXMessageError(f"msg_type=`{msg.msg_type}` not in schema")

        schema_msg = self._messages_types[msg.msg_type]

        schema_fields = set()
        for fname, req in schema_msg.required.items():
            f = schema_msg[fname]
            schema_fields.add(fname)

            if req:
                if isinstance(f, SchemaField):
                    if f.tag not in msg:
                        raise FIXMessageError(f"Missing required field={repr(f)}")

        if "8" in msg:
            self._validate_header(msg)

        for tag, val in msg.tags.items():
            if tag == "10":
                # TODO: check the checksum
                continue
            if tag not in self._tag2field:
                raise FIXMessageError(f"msg tag={tag} not in schema")
            field = self._tag2field[tag]

            if field in self._header:
                continue

            if field not in schema_msg:
                raise FIXMessageError(
                    f"msg field={field} is not allowed in {schema_msg}"
                )

            fschema = schema_msg[field]
            if isinstance(fschema, SchemaField):
                if msg.is_group(tag):
                    raise FIXMessageError(
                        f"msg tag={tag} val={val} must be a tag, got group"
                    )
                fschema.validate_value(val)

            elif isinstance(fschema, SchemaGroup):
                if not msg.is_group(tag):
                    raise FIXMessageError(f"msg tag={tag} val={val} must be a group")
                fschema.validate_group(msg.get_group_list(tag))

        return True

    def __getitem__(self, item: int | str | FTag) -> SchemaField:
        """Get SchemaField by tag."""
        try:
            tag = str(int(str(item)))
            return self._tag2field[tag]
        except ValueError:
            return self._field2tag[str(item)]



"""FIX Message encoding / decoding module."""
import logging
from datetime import datetime

from asyncfix import FMsg, FTag
from asyncfix.errors import EncodingError
from asyncfix.message import FIXContainer, FIXMessage, RepeatingTagError
from asyncfix.protocol import FIXProtocolBase
from asyncfix.session import FIXSession


class _RepeatingGroupContext(FIXContainer):
    def __init__(self, tag, repeating_group_tags, parent):
        self.tag = tag
        self.repeating_group_tags = repeating_group_tags
        self.parent = parent
        FIXContainer.__init__(self)


class Codec(object):
    """Encoding / decoding engine.

    Attributes:
        protocol: FIX protocol
        SOH: encoded message separator
    """

    def __init__(self, protocol: FIXProtocolBase):
        """Codec init.

        Args:
            protocol: FIX protocol used in encoding/decoding
        """
        self.protocol: FIXProtocolBase = protocol
        self.SOH = "\x01"

    @staticmethod
    def current_datetime() -> str:
        """FIX complaint date-time string (UTC now)."""
        return datetime.utcnow().strftime("%Y%m%d-%H:%M:%S.%f")[:-3]

    def _addTag(self, body, t, msg: FIXMessage):
        if msg.is_group(t):
            groups = msg.get_group_list(t)
            body.append("%s=%s" % (t, len(groups)))
            for group in groups:
                for tag in group.tags:
                    self._addTag(body, tag, group)
        else:
            body.append("%s=%s" % (t, msg[t]))

    def encode(
        self,
        msg: FIXMessage,
        session: FIXSession,
        raw_seq_num: bool = False,
    ) -> str:
        """Encodes FIXMessage into serialized message.

        Args:
            msg: generic FIXMessage
            session: current session (for seq num)
            raw_seq_num: if True - uses MsgSeqNum from `msg`

        Returns:
            encoded message (string)

        Raises:
            EncodingError: when failed MsgSeqNum conditions for some types of messages
        """
        # Create body
        body = []

        msg_type = msg.msg_type

        body.append("%s=%s" % (FTag.SenderCompID, session.sender_comp_id))
        body.append("%s=%s" % (FTag.TargetCompID, session.target_comp_id))

        seq_no = 0
        if raw_seq_num:
            seq_no = msg[FTag.MsgSeqNum]
        else:
            if msg_type == FMsg.SEQUENCERESET:
                if FTag.MsgSeqNum not in msg:
                    raise EncodingError(
                        "SequenceReset must have the MsgSeqNum already populated"
                    )
                seq_no = msg[FTag.MsgSeqNum]
            else:
                # if we have the PossDupFlag set, we need to send the message
                #   with the same seqNo
                if msg.get(FTag.PossDupFlag, "N") == "Y":
                    if FTag.MsgSeqNum not in msg:
                        raise EncodingError(
                            "Failed to encode message with PossDupFlag=Y but no"
                            " previous MsgSeqNum"
                        )
                    seq_no = msg[FTag.MsgSeqNum]
                else:
                    seq_no = session.allocate_next_num_out()

        body.append("%s=%s" % (FTag.MsgSeqNum, seq_no))
        body.append("%s=%s" % (FTag.SendingTime, self.current_datetime()))

        for t in msg.tags:
            if t in {
                FTag.MsgSeqNum,
                FTag.SendingTime,
                FTag.SenderCompID,
                FTag.TargetCompID,
            }:
                continue
            self._addTag(body, t, msg)

        # Enable easy change when debugging
        SEP = self.SOH

        body = self.SOH.join(body) + self.SOH

        # Create header
        header = []
        msg_type = "%s=%s" % (FTag.MsgType, msg_type)
        header.append("%s=%s" % (FTag.BeginString, self.protocol.beginstring))
        header.append("%s=%i" % (FTag.BodyLength, len(body) + len(msg_type) + 1))
        header.append(msg_type)

        fixmsg = self.SOH.join(header) + self.SOH + body

        cksum = sum([ord(i) for i in fixmsg]) % 256
        fixmsg = fixmsg + "%s=%0.3i" % (FTag.CheckSum, cksum)

        # print len(fixmsg)

        return fixmsg + SEP

    def decode(
        self,
        rawmsg: bytes,
        silent: bool = True,
    ) -> tuple[FIXMessage | None, int, bytes | None]:
        """Decodes message from socket.

        Args:
            rawmsg: message bytes
            silent: no errors raised, returns non

        Returns:
            if OK - (FIXMessage, bytes_processed, valid_raw_msg_bytes)
            if ERR - (None, n_bytes_skip, None)
        """
        valid_idx = rawmsg.find(b"8=FIX.")
        if valid_idx == -1:
            assert silent, "no fix header"
            return None, len(rawmsg), None

        parsed_length = valid_idx

        msg = rawmsg[valid_idx:].decode("latin-1")

        next_msg = msg[5:].find("8=FIX.")
        if next_msg != -1:
            # Next fix message added, but incomplete
            next_msg += 5
        else:
            next_msg = len(msg)

        encoded_msg = rawmsg[valid_idx : next_msg + valid_idx]

        msg = msg[:next_msg].split(self.SOH)
        if not msg[-1]:
            msg = msg[:-1]

        # at a minimum we require BeginString, BodyLength & Checksum
        if len(msg) < 3:
            assert silent, "Minimum message"
            return (None, parsed_length, None)

        tag, value = msg[0].split("=", 1)
        if value != self.protocol.beginstring:
            logging.error(
                "FIX Version unexpected (Recv: %s Expected: %s)"
                % (value, self.protocol.beginstring)
            )
            assert silent, "protocol beginstring mismatch"
            return (None, len(rawmsg), None)

        toks = msg[1].split("=", 1)
        if len(toks) != 2:
            assert silent, f"BodyLength split error {msg}"
            return (None, len(rawmsg), None)
        tag, value = toks

        msg_length = len(msg[0]) + len(msg[1]) + len("10=000") + 3
        if tag != FTag.BodyLength:
            logging.error(f"*** BodyLength missing or not 2nd field *** [{tag}]: {msg}")
            assert silent, "2nd tag must be BodyLength"
            return (None, len(rawmsg), None)
        else:
            msg_length += int(value)

        # message looks incomplete
        if msg_length > len(rawmsg):
            assert silent, "incomplete message"
            return (None, parsed_length, None)

        checksum_passed = False
        parsed_length += msg_length

        decoded_msg = FIXMessage("UNKNOWN")
        repeating_groups = []
        repeating_group_tags = self.protocol.repeating_groups
        current_context = decoded_msg

        for m in msg:
            toks = m.split("=", 1)
            if len(toks) != 2:
                assert silent, f"incomplete tag {m}"
                return (None, len(rawmsg), None)
            tag, value = toks

            if tag == FTag.CheckSum:
                cheksum_base = self.SOH.join(msg[:-1])
                checksum = (sum([ord(i) for i in cheksum_base]) + 1) % 256

                if checksum != int(value):
                    logging.warning(
                        "\tCheckSum: %s (INVALID) expecting %s" % (int(value), checksum)
                    )
                    assert (
                        silent
                    ), f"invalid checksum tag[10]={value} expected: {checksum} {msg=}"
                    checksum_passed = False
                else:
                    checksum_passed = True
            elif tag == FTag.MsgType:
                try:
                    decoded_msg.msg_type = FMsg(value)
                except ValueError:
                    decoded_msg.msg_type = value

            # found the start of a repeating group
            if tag in repeating_group_tags:
                # i.e. we are already in a repeating group
                if type(current_context) is _RepeatingGroupContext:
                    while (
                        repeating_groups
                        and tag not in current_context.repeating_group_tags
                    ):
                        current_context.parent.add_group(
                            current_context.tag, current_context
                        )
                        current_context = current_context.parent
                        # pop the completed group off the stack
                        del repeating_groups[-1]

                ctx = _RepeatingGroupContext(
                    tag, repeating_group_tags[tag], current_context
                )
                repeating_groups.append(ctx)
                current_context = ctx
            elif repeating_groups:
                # we have 1 or more repeating groups in progress
                #    & our tag isn't the start of a group
                while (
                    repeating_groups and tag not in current_context.repeating_group_tags
                ):
                    current_context.parent.add_group(
                        current_context.tag, current_context
                    )
                    current_context = current_context.parent
                    # pop the completed group off the stack
                    del repeating_groups[-1]

                if tag in current_context.tags:
                    # if the repeating group already contains this field,
                    #     start the next
                    current_context.parent.add_group(
                        current_context.tag, current_context
                    )
                    ctx = _RepeatingGroupContext(
                        current_context.tag,
                        current_context.repeating_group_tags,
                        current_context.parent,
                    )
                    del repeating_groups[-1]
                    repeating_groups.append(ctx)
                    current_context = ctx

                # else add it to the current one
                current_context.set(tag, value)
            else:
                if tag in decoded_msg:
                    # Repeating tag found, possibly RepGrp not in protocol schema
                    decoded_msg.set(tag, RepeatingTagError)
                else:
                    # this isn't a repeating group field, so just add it normally
                    decoded_msg.set(tag, value)

        if checksum_passed:
            return (decoded_msg, parsed_length, encoded_msg)
        else:
            assert silent, f"Checksum probably missing: {msg}"
            return (None, parsed_length, None)
